// includes
#include <OneWire.h>                // for Dallas temperature sensors
#include <DallasTemperature.h>      // for Dallas temperature sensors
#include <avr/wdt.h>                // for reset via watchdog

// constants
const int oneWirePin                = 2;
const int pwmPins[]                 = {5, 9, 10, 11};
const int possibleTachoPins[]       = {4, 7, 8, 12};
const float tempPrecisionInDegree   = 0.25;         // 0.5|0.25|0.125|0.0625
const float minTempForMinRPM        = 30.0;         // fans starts to speed up            
const float minTempForMaxRPM        = 37.5;         // fans run at full speed
const int rpmDifferenceThreshold    = 25;
const int measureIntervalInSeconds  = 5;

// variables
int numberOfTemperatureSensors      = 0;
int tachoPins[]                     = {-1, -1, -1, -1};
int minPWMs[]                       = {0, 0, 0, 0};
int pwmValues[]                     = {-1, -1, -1, -1};
int minRPMs[]                       = {0, 0, 0, 0};
int maxRPMs[]                       = {0, 0, 0, 0};
float lastTemperature               = 0;
float lastMeasure                   = millis();

// pre setup initialization
OneWire oneWire(oneWirePin);
DallasTemperature sensors(&oneWire);

/*
 * Function:  softReset 
 * --------------------
 *   Set watchdog timer for automatic reset after given ms.
 *
 *  msToReset: number of ms till the watchdog resets the arduino.
 */
void softReset(uint8_t milliesecondsBeforeReset)
{
    wdt_enable(milliesecondsBeforeReset); // start watchdog
    while(true){};   // wait without sending anything, so wdt_reset() is called.
}

/*
 * Function:  setPWM 
 * --------------------
 *   Sets the PWM value to given pin:
 *
 *   pin: pin number, the pwm value will be set. valid values are: 5|9|10|11.
 *   val: pwm value that will be set.
 *
 *   Any value <0 will set to 0 and every value >100 will be set to 100.
 */
void setPWM(int pin, int val)
{
    // Serial.print("set PWM called with pin ");
    // Serial.print(pin); Serial.print(" value "); Serial.println(val);

    val = (val<0) ? 0 : (val>100) ? 100 : val;  // only 0-100 is allowed
    int value = ((val*320)/100);  // re-calc to 0-320

    switch(pin)
    {
        case  5: OCR3A = value; break;
        case  9: OCR1A = value; break;
        case 10: OCR1B = value; break;
        case 11: OCR1C = value; break;
    }
}

/*
 * Function:  getHighestTemperature 
 * --------------------
 *   Reads all temperature sensors and returns the highest temperature.
 *
 *   returns: the highest measured temperature
 *            returns minTempForMaxRPM if numberOfTemperatureSensors == 0
 */
float getHighestTemperature()
{
    // command to get temperatures
    sensors.requestTemperatures(); 

    float highestTempC = -1;
    for (int i=0 ; i<numberOfTemperatureSensors ; i++)
    {
        float tempC = sensors.getTempCByIndex(i);
        highestTempC = (highestTempC<tempC) ? tempC : highestTempC;
    }

    return (highestTempC==-1) ?  minTempForMaxRPM : highestTempC;
}

/*
 * Function: setupTemperatureSensors
 * ----------------------------
 *   Detect the number of DS18B20 Sensors and set the temperature precision.
 *
 *   Requests a few temperature values to get usable values.
 */
void setupTemperatureSensors()
{
    while(true)
    {
        sensors.begin();
        numberOfTemperatureSensors = sensors.getDS18Count();
        // Serial.print("found "); Serial.print(numberOfTemperatureSensors);
        // Serial.println(" DS18b20 sensors.");

        if (numberOfTemperatureSensors>0) break;

        Serial.print("need at least one DS18b20 sensor.");
        Serial.println(" none found. check wires. retry in 5s");
        delay(5000);
    }

    if (tempPrecisionInDegree == 0.0625) 
    {
        sensors.setResolution(0x7F);
    } 
    else if (tempPrecisionInDegree == 0.125)
    {
        sensors.setResolution(0x5F);
    }
    else if (tempPrecisionInDegree == 0.25)
    {
        sensors.setResolution(0x3F);
    }
    else
    {
        sensors.setResolution(0x1F);
    }

    Serial.print("Found "); Serial.print(numberOfTemperatureSensors); 
    Serial.println(" x DS18b20 sensors.");
    
    // first values mostly crap, so we request a few times
    sensors.requestTemperatures();
    delay(500);
    sensors.requestTemperatures();
}

/*
 * Function: measureRPM
 * ----------------------------
 *   Measures the rotation per minute of the fan at the over given pin.
 *  
 *
 *   pinNumber: has to be between 0 and 13.
 *   roundedBy: round up the value, default is 10.
 *
 *   returns: RPM of the fan at the over given pin. 
 */
int measureRPM(int pinNumber, int roundedBy = 10)
{
    if (pinNumber<0 || pinNumber>13) return 0;
    unsigned long edgeTime = pulseIn(pinNumber, LOW);
    int rpm = ((1000/((edgeTime * 4)/1000))*6*10);
    // Serial.print("tacho on pin "); Serial.print(pinNumber); 
    // Serial.print(" has "); Serial.print(rpm); Serial.println(" RPM");
    rpm = (rpm<0) ? 0 : rpm;
    return ((rpm / roundedBy) * roundedBy);
}

/*
 * Function: foudAtleastOneFan
 * ----------------------------
 *   Returns true if one or more fans where found at detectFans-function.
 *
 *   returns: true it one or more fans connected, false otherwise. 
 */
bool foudAtleastOneFan()
{
    bool missing1 = (tachoPins[0] == -1);
    bool missing2 = (tachoPins[1] == -1);
    bool missing3 = (tachoPins[2] == -1);
    bool missing4 = (tachoPins[3] == -1);
    bool noneFans = (missing1 && missing2 && missing3 && missing4);
    return !noneFans;
}

/*
 * Function: detectFans
 * ----------------------------
 *   Searchs all pins for fans.
 */
void detectFans()
{
    while(true)
    {
        Serial.println("checking tacho pins for signals...");

        // Serial.println("set all fans pwm to 0.");
        for (int fan=0 ; fan<4 ; fan++) setPWM(pwmPins[fan],0);
        delay(3000); // give the fans time to slow down.

        // collect minimal rpm for all pins
        // Serial.println("collect min RPMs");
        // problem: measuring is not 100 correct. we need more than one try.
        for (int tryNumber=0 ; tryNumber<3 ; tryNumber++)
        {
            for (int fan=0 ; fan<4 ; fan++)
            {
                int measureValue = measureRPM(possibleTachoPins[fan]);
                minRPMs[fan] = (minRPMs[fan] < measureValue) ? measureValue : minRPMs[fan];
            }
        }
        
        // speed up fan by fan, measure RPM difference (highest=responding fan)
        for (int fan=0 ; fan<4 ; fan++)
        {
            // Serial.print("speed up on pin "); Serial.println(pwmPins[fan]);
            setPWM(pwmPins[fan], 100);
            delay(4000);

            // Serial.println("search for max differ rpm.");
            int maxDiff = 0;
            int key = 0;
            int uniqueDiff = true;
            for (int tacho=0 ; tacho<4 ; tacho++)
            {
                int rpm = measureRPM(possibleTachoPins[tacho]);
                int diff = rpm - minRPMs[tacho];
                // Serial.print("old minRPMs[tacho] was: "); 
                // Serial.println(minRPMs[tacho]); Serial.print("slot "); 
                // Serial.print(tacho); Serial.print(" is pin ");
                // Serial.print(possibleTachoPins[tacho]); 
                // Serial.print(" has "); Serial.print(rpm); 
                // Serial.print(" RPM");  Serial.print(" differs ");
                // Serial.println(diff);

                if (diff <= rpmDifferenceThreshold) continue; // skip non differs
              
                if (diff > maxDiff)
                {
                    maxDiff = diff;
                    key = tacho;
                }
                else if (diff == maxDiff) uniqueDiff = false;
            }

            // filter results
            if (maxDiff > 0 && uniqueDiff)
            {
                bool unique = true;
                for(int i=0; i<4 ; i++) if (tachoPins[i]==possibleTachoPins[key]) unique = false;
                if(unique) 
                {
                    tachoPins[fan] = possibleTachoPins[key]; // save pin
                    maxRPMs[fan] = minRPMs[fan] + maxDiff;
                    Serial.print("found: fan"); Serial.print(fan);
                    Serial.print(" pwm_pin="); Serial.print(pwmPins[fan]);
                    Serial.print(" tacho_pin="); Serial.print(tachoPins[fan]);
                    Serial.print(" has min_rpm="); Serial.print(minRPMs[fan]);
                    Serial.print(" RPM, at 100 pwm it has ");
                    Serial.print(maxDiff); Serial.println(" more RPM.");
                }
            }
            setPWM(pwmPins[fan], 0);
        }

        if (foudAtleastOneFan()) return;
        Serial.println("no fans found. need at least one fan. re-checking in 5s.");
        delay(5000);
    }
}

/*
 * Function: measureMinPWMForFans
 * ----------------------------
 *   Detect for each Fan the highest PWM for the lowest RPM greater than zero.
 */
void measureMinPWMForFans()
{
    Serial.println("starting fan check, may take a bit...");

    for (int fan=0 ; fan<4 ; fan++)
    {
        if (tachoPins[fan] < 0) continue; // skip not existing fans
        // Serial.print("slowing down fan "); Serial.print(fan); 
        // Serial.print(" at pin "); Serial.println(pwmPins[fan]);
        setPWM(pwmPins[fan], 0);
        delay(2000);
        
        int minRPM = minRPMs[fan];
        int minPWM = 0;
        // Serial.println("speeding up the fan, step by step.");

        for (int pwm = 0; pwm <= 100; pwm = pwm + 5)
        {
            // Serial.print("set at fan index "); Serial.print(fan);
            // Serial.print(" which has tach on pin");
            // Serial.print(tachoPins[fan]); Serial.print(" to pwm value "); 
            // Serial.println(pwm);
            setPWM(pwmPins[fan], pwm);
            delay(1000);
            int rpm = measureRPM(tachoPins[fan], 100);
            // Serial.print("after 1 sec i measure "); 
            // Serial.print(rpm); Serial.println(" RPM");
      
            // the check for 'rpm < minRPM' fixes the case than one measure 
            // get lower rpm than minimal rpm. rpm vary from time to time.
            if (rpm == minRPM || rpm < minRPM) 
            {
                minPWM = pwm; // found a higher pwm for same rpm
            }
            else break; // rpm raised, lowest pwm found, skip here
        }

        setPWM(pwmPins[fan], 0);
        minPWMs[fan] = minPWM; // save value
        Serial.print("Mininal RPM for fan"); Serial.print(fan);
        Serial.print(" is "); Serial.print(minRPM);
        Serial.print(" at PWM value "); Serial.println(minPWM);
    }
}

/*
 * Function: configureTimer
 * ----------------------------
 *   Configures both 16bit timer for 25 kHz pwm signals.
 */
void configureTimer()
{
    // Configure Timer 1 for PWM @ 25 kHz.
    TCCR1A = 0;   // undo the configuration done by...
    TCCR1B = 0;   // ...the Arduino core library...
    TCCR1C = 0;   // ...for pin 9, 10 and 11.
    TCNT1  = 0;   // reset timer
    // non-inverted PWM on ch. A, B and C, mode 10: ph. correct PWM, TOP = ICR1
    TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(COM1C1) | _BV(WGM11);
    TCCR1B = _BV(WGM13) | _BV(CS10); // prescaler = 1
    TCCR1C = _BV(WGM13) | _BV(CS10); // on PIN 9,10,11
    ICR1   = 320;       // TOP = 320, @16Mhz CPU -> 25kHz PWM
    // Configure Timer 3
    TCCR3A = 0;
    TCNT3  = 0;
    TCCR3A = _BV(COM1A1) | _BV(WGM11);
    TCCR3B = _BV(WGM13) | _BV(CS10);
    ICR3   = 320;
}

/*
 * Function: showConfig
 * ----------------------------
 *   Prints the current pwm, rpm and temperatures and limits for all fans.
 *
 *   compact: if true, it prints in nagios perf data format.
 *            if false, it prints full text format.
 */
void showConfig(bool compact = false)
{
    if (compact)
    {
        // nagios perf data format: 'label'=value[UOM];[warn];[crit];[min];[max]
        String pwm = "", rpm = "";
        for (int fan=0 ; fan<4 ; fan++)
        {
            pwm = pwm + "pwm" + fan + "=" + ((pwmValues[fan]>0)?pwmValues[fan]:0) + ";;;0;100 ";
            rpm = rpm + "rpm" + fan + "=" + measureRPM(tachoPins[fan]) + ";;" + minRPMs[fan] + ";" + maxRPMs[fan] + "; ";
        }
        String tmp = "";
        for (int i=0 ; i<5 ; i++)
        {
            float tempC = sensors.getTempCByIndex(i);
            tmp = tmp + "tmp" + i + "=" + ((tempC>0)?tempC:0) + ";;;;; ";
        }
        Serial.println(pwm + rpm + tmp);
    }
    else
    {
        Serial.println("--- current config ---");
        for (int fan=0 ; fan<4 ; fan++)
        {
            if (tachoPins[fan] == -1)
            {
                Serial.print("no tacho found for fan "); Serial.print(fan);
                Serial.print(" at pin  "); Serial.println(pwmPins[fan]);
            }
            else
            {
                Serial.print("tacho for fan "); Serial.print(fan);
                Serial.print(" at pin "); Serial.print(pwmPins[fan]);
                Serial.print(" was found on pin "); Serial.print(tachoPins[fan]);
                Serial.print(" the minimal (with rpm>0) pwm value "); Serial.println(minPWMs[fan]);
            }
        }
        Serial.println("--- current config end ---");
    }
}

/*
 * Function: setup
 * ----------------------------
 *   Default arduino setup function, calls all the initialization functions.
 */
void setup()
{
    Serial.begin(9600);     // initialization of serial over usb.
    delay(2000);            // wait a bit for external program to connect.
    Serial.println("fan control setup start.");

    // set pin into needed pin-mode
    Serial.println("set input pins...");
    for (int i=0 ; i<4 ; i++) pinMode(possibleTachoPins[i], INPUT);
    Serial.println("set output pins...");
    for (int i=0 ; i<4 ; i++) pinMode(pwmPins[i], OUTPUT);

    // configure timer
    Serial.println("configure timer for 25kHz PWM...");
    configureTimer();

    // slow down all fans
    for (int fan=0 ; fan<4 ; fan++) setPWM(pwmPins[fan],0);

    setupTemperatureSensors();
    detectFans();
    measureMinPWMForFans();
    showConfig();

    // set fans to minmal pwms (every temperature change will alter this)
    for (int fan=0 ; fan<4 ; fan++) setPWM(pwmPins[fan], minPWMs[fan]);

    Serial.println("fan control setup done.");
}

/*
 * Function: loop
 * ----------------------------
 *   Default arduino main loop.
 */
void loop()
{
    if (millis() > (lastMeasure + (measureIntervalInSeconds*1000)))
    {
        // new measure interval started
        // Serial.print("measure interval started at "); Serial.println(millis());
        float currentTemperature = getHighestTemperature();

        if (currentTemperature != lastTemperature)
        {
            // Serial.print("temp changed from "); Serial.print(lastTemperature);
            // Serial.print(" to "); Serial.println(currentTemperature);
            for (int fan=0 ; fan<4 ; fan++)
            {
                if (tachoPins[fan] < 0) continue; // skip not existing fans
                // Serial.print("for fan index "); Serial.println(fan);

                int newPWM = -1;
                if (currentTemperature<minTempForMinRPM)
                {
                    newPWM = minPWMs[fan];
                }
                else if (currentTemperature>=minTempForMaxRPM)
                {
                    newPWM = 100;
                }
                else
                {
                    // calc new pwm value
                    // Serial.println("pwm has to be calculated");
                    int steps = ((minTempForMaxRPM - minTempForMinRPM) / tempPrecisionInDegree) + 1;
                    // Serial.print("steps "); Serial.println(steps);
                    int pwmRange = 100 -  minPWMs[fan];
                    // Serial.print("pwm range "); Serial.println(pwmRange);
                    int pwmIncPerStep = pwmRange / steps;
                    // Serial.print("pwmIncPerStep "); Serial.println(pwmIncPerStep);
                    int stepsToGo = ((currentTemperature - minTempForMinRPM) / tempPrecisionInDegree) + 1;
                    // Serial.print("stepsToGo "); Serial.println(stepsToGo);
                    newPWM = minPWMs[fan] + (stepsToGo * pwmIncPerStep);
                }

                // set the pwm if it changed
                // Serial.print("new pwm value should be "); Serial.println(newPWM);
                if (newPWM != pwmValues[fan])
                {
                    //Serial.print("!!!!! Differs, so will be set to "); Serial.println(newPWM);
                    pwmValues[fan] = newPWM;
                    setPWM(pwmPins[fan],newPWM);
                }
            }
            lastTemperature = currentTemperature;
        }
        lastMeasure = millis();
    }

    if (Serial.available() > 0)
    {
        // flush input buffer
        while(Serial.available() > 0)
        {
            char t = Serial.read();
            if (t=='r') softReset(100);
        }
        showConfig(true);
    }
}
